//
//  StatusCache.swift
//  Tusker
//
//  Created by Shadowfacts on 9/17/18.
//  Copyright © 2018 Shadowfacts. All rights reserved.
//

import Foundation
import Combine
import Pachyderm

class MastodonCache {
    
    private static var statuses = CachedDictionary<Status>(name: "Statuses")
    private static var accounts = CachedDictionary<Account>(name: "Accounts")
    private static var relationships = CachedDictionary<Relationship>(name: "Relationships")
    private static var notifications = CachedDictionary<Pachyderm.Notification>(name: "Notifications")
    
    static let statusSubject = PassthroughSubject<Status, Never>()
    static let accountSubject = PassthroughSubject<Account, Never>()

    // MARK: - Statuses
    static func status(for id: String) -> Status? {
        return statuses[id]
    }
    
    static func set(status: Status, for id: String) {
        statuses[id] = status
        add(account: status.account)
        if let reblog = status.reblog {
            add(status: reblog)
            add(account: reblog.account)
        }
        
        statusSubject.send(status)
    }
    
    static func status(for id: String, completion: @escaping (Status?) -> Void) {
        let request = Client.getStatus(id: id)
        MastodonController.run(request) { response in
            guard case let .success(status, _) = response else {
                completion(nil)
                return
            }
            set(status: status, for: id)
            completion(status)
        }
    }
    
    static func add(status: Status) {
        set(status: status, for: status.id)
    }
    
    static func addAll(statuses: [Status]) {
        statuses.forEach(add)
    }
    
    // MARK: - Accounts
    static func account(for id: String) -> Account? {
        return accounts[id]
    }
    
    static func set(account: Account, for id: String) {
        accounts[id] = account
        accountSubject.send(account)
    }
    
    static func account(for id: String, completion: @escaping (Account?) -> Void) {
        let request = Client.getAccount(id: id)
        MastodonController.run(request) { response in
            guard case let .success(account, _) = response else {
                completion(nil)
                return
            }
            set(account: account, for: account.id)
            completion(account)
        }
    }
    
    static func add(account: Account) {
        set(account: account, for: account.id)
    }
    
    static func addAll(accounts: [Account]) {
        accounts.forEach(add)
    }
    
    // MARK: - Relationships
    static func relationship(for id: String) -> Relationship? {
        return relationships[id]
    }
    
    static func set(relationship: Relationship, id: String) {
        relationships[id] = relationship
    }
    
    static func relationship(for id: String, completion: @escaping (Relationship?) -> Void) {
        let request = Client.getRelationships(accounts: [id])
        MastodonController.run(request) { response in
            guard case let .success(relationships, _) = response,
                let relationship = relationships.first else {
                completion(nil)
                return
            }
            set(relationship: relationship, id: relationship.id)
            completion(relationship)
        }
    }
    
    static func add(relationship: Relationship) {
        set(relationship: relationship, id: relationship.id)
    }
    
    static func addAll(relationships: [Relationship]) {
        relationships.forEach(add)
    }
    
    // MARK: - Notifications
    static func notification(for id: String) -> Pachyderm.Notification? {
        return notifications[id]
    }

    static func set(notification: Pachyderm.Notification, id: String) {
        notifications[id] = notification
    }
    
    static func add(notification: Pachyderm.Notification) {
        set(notification: notification, id: notification.id)
    }
    
    static func addAll(notifications: [Pachyderm.Notification]) {
        notifications.forEach(add)
    }
    
}

class CachedDictionary<Value> {
    private let name: String
    private var dict = [String: Value]()
    private let queue: DispatchQueue
    
    init(name: String) {
        self.name = name
        self.queue = DispatchQueue(label: "CachedDictionary (\(name)) Coordinator", attributes: .concurrent)
    }
    
    subscript(key: String) -> Value? {
        get {
            var result: Value? = nil
            queue.sync {
                result = dict[key]
            }
            return result
        }
        set(value) {
            queue.async(flags: .barrier) {
                self.dict[key] = value
            }
        }
    }
}
